/*
 * Copyright 2013 Mark Slater
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at
 *
 * 	http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 */

package net.sourceforge.xazzle.xhtml;

import org.sourceforge.writexml.Attribute;
import org.sourceforge.writexml.Attributes;
import org.sourceforge.writexml.TagName;

import static java.util.Arrays.asList;
import static net.sourceforge.xazzle.xhtml.IterableUtils.*;
import static org.sourceforge.writexml.Attribute.attribute;
import static org.sourceforge.writexml.Attributes.attributes;
import static org.sourceforge.writexml.TagName.tagName;

public final class TableTag<T extends MayContainFormTag> extends StandardXhtmlTag<TableTag<T>, XhtmlTag> implements XhtmlTag, BodyElement<T, TagScope.Block> {

    private TableTag(final TagName tagName, final Attributes attributes, final Iterable<? extends XhtmlTag> children) {
        super(tagName, attributes, children);
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final TableRowTag<? extends T> tableRowTag, final TableRowTag<? extends T>... tableRowTags) {
        return tableTag(tableRowTag, asList(tableRowTags));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final TableRowTag<? extends T> tableRowTag, final Iterable<TableRowTag<? extends T>> tableRowTags) {
        return makeATableTagWithAtLeastOneChild(cons(tableRowTag, tableRowTags));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final CaptionTag captionTag, final TableRowTag<? extends T> tableRowTag, final TableRowTag<? extends T>... tableRowTags) {
        return tableTag(captionTag, tableRowTag, asList(tableRowTags));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final CaptionTag captionTag, final TableRowTag<? extends T> tableRowTag, final Iterable<TableRowTag<? extends T>> tableRowTags) {
        return TableTag.makeATableTagWithAtLeastOneChild(cons(captionTag, cons(tableRowTag, tableRowTags)));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final TableBodyTag<? extends T> tableBodyTag) {
        return makeATableTagWithAtLeastOneChild(asList(tableBodyTag));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final CaptionTag captionTag, final TableBodyTag<? extends T> tableBodyTag) {
        return makeATableTagWithAtLeastOneChild(asList(captionTag, tableBodyTag));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final TableHeaderTag<? extends T> tableHeaderTag, final TableRowTag<? extends T> tableRowTag, final TableRowTag<? extends T>... tableRowTags) {
        return makeATableTagWithAtLeastOneChild(concat(asList(tableHeaderTag, tableRowTag), asList(tableRowTags)));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final TableHeaderTag<? extends T> tableHeaderTag, final TableRowTag<? extends T> tableRowTag, final Iterable<TableRowTag<? extends T>> tableRowTags) {
        return makeATableTagWithAtLeastOneChild(concat(asList(tableHeaderTag, tableRowTag), tableRowTags));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final CaptionTag captionTag, final TableHeaderTag<? extends T> tableHeaderTag, final TableRowTag<? extends T> tableRowTag, final TableRowTag<? extends T>... tableRowTags) {
        return makeATableTagWithAtLeastOneChild(concat(asList(captionTag, tableHeaderTag, tableRowTag), asList(tableRowTags)));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final TableHeaderTag<? extends T> tableHeaderTag, final TableBodyTag<? extends T> tableBodyTag) {
        return makeATableTagWithAtLeastOneChild(asList(tableHeaderTag, tableBodyTag));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final CaptionTag captionTag, final TableHeaderTag<? extends T> tableHeaderTag, final TableBodyTag<? extends T> tableBodyTag) {
        return makeATableTagWithAtLeastOneChild(asList(captionTag, tableHeaderTag, tableBodyTag));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final ColumnSpecifications columnSpecifications, final TableRowTag<? extends T> tableRowTag, final TableRowTag<? extends T>... tableRowTags) {
        return tableTag(columnSpecifications, tableRowTag, asList(tableRowTags));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final ColumnSpecifications columnSpecifications, final TableRowTag<? extends T> tableRowTag, final Iterable<TableRowTag<? extends T>> tableRowTags) {
        return makeATableTagWithAtLeastOneChild(
                concat(
                        append(columnSpecifications, (XhtmlTag) tableRowTag),
                        tableRowTags
                ));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final CaptionTag captionTag, final ColumnSpecifications columnSpecifications, final TableRowTag<? extends T> tableRowTag, final TableRowTag<? extends T>... tableRowTags) {
        return tableTag(captionTag, columnSpecifications, tableRowTag, asList(tableRowTags));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final CaptionTag captionTag, final ColumnSpecifications columnSpecifications, final TableRowTag<? extends T> tableRowTag, final Iterable<TableRowTag<? extends T>> tableRowTags) {
        return makeATableTagWithAtLeastOneChild(
                concat(
                        append(cons((XhtmlTag) captionTag, columnSpecifications), tableRowTag),
                        tableRowTags
                ));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final ColumnSpecifications columnSpecifications, final TableBodyTag<? extends T> tableBodyTag) {
        return makeATableTagWithAtLeastOneChild(append(columnSpecifications, (XhtmlTag) tableBodyTag));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final CaptionTag captionTag, final ColumnSpecifications columnSpecifications, final TableBodyTag<? extends T> tableBodyTag) {
        return makeATableTagWithAtLeastOneChild(append(cons((XhtmlTag) captionTag, columnSpecifications), tableBodyTag));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final ColumnSpecifications columnSpecifications, final TableHeaderTag<? extends T> tableHeaderTag, final TableRowTag<? extends T> tableRowTag, final TableRowTag<? extends T>... tableRowTags) {
        return tableTag(columnSpecifications, tableHeaderTag, tableRowTag, asList(tableRowTags));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final ColumnSpecifications columnSpecifications, final TableHeaderTag<? extends T> tableHeaderTag, final TableRowTag<? extends T> tableRowTag, final Iterable<TableRowTag<? extends T>> tableRowTags) {
        return makeATableTagWithAtLeastOneChild(
                concat(
                        append(append(columnSpecifications, (XhtmlTag) tableHeaderTag), tableRowTag),
                        tableRowTags
                ));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final CaptionTag captionTag, final ColumnSpecifications columnSpecifications, final TableHeaderTag<? extends T> tableHeaderTag, final TableRowTag<? extends T> tableRowTag, final TableRowTag<? extends T>... tableRowTags) {
        return tableTag(captionTag, columnSpecifications, tableHeaderTag, tableRowTag, asList(tableRowTags));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final CaptionTag captionTag, final ColumnSpecifications columnSpecifications, final TableHeaderTag<? extends T> tableHeaderTag, final TableRowTag<? extends T> tableRowTag, final Iterable<TableRowTag<? extends T>> tableRowTags) {
        return makeATableTagWithAtLeastOneChild(
                concat(
                        append(append(cons((XhtmlTag) captionTag, columnSpecifications), tableHeaderTag), tableRowTag),
                        tableRowTags
                ));
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final ColumnSpecifications columnSpecifications, final TableHeaderTag<? extends T> tableHeaderTag, final TableBodyTag<? extends T> tableBodyTag) {
        return makeATableTagWithAtLeastOneChild(
                append(append(columnSpecifications, (XhtmlTag) tableHeaderTag), tableBodyTag)
        );
    }

    static <T extends MayContainFormTag> TableTag<T> tableTag(final CaptionTag captionTag, final ColumnSpecifications columnSpecifications, final TableHeaderTag<? extends T> tableHeaderTag, final TableBodyTag<? extends T> tableBodyTag) {
        return makeATableTagWithAtLeastOneChild(append(append(cons((XhtmlTag) captionTag, columnSpecifications), tableHeaderTag), tableBodyTag));
    }

    private static <T extends MayContainFormTag> TableTag<T> makeATableTagWithAtLeastOneChild(final Iterable<? extends XhtmlTag> mandatoryTags) {
        return new TableTag<>(tagName("table"), attributes(), mandatoryTags);
    }

    @Override
    TableTag<T> makeANewOne(final Attribute attribute) {
        return new TableTag<>(tagName, attributes.append(attribute), children);
    }

    public TableTag<T> withBorder(final Border border) {
        return makeANewOne(attribute("border", border, NAMESPACE));
    }

    public TableTag<T> withCellPadding(final CellPadding cellPadding) {
        return makeANewOne(attribute("cellpadding", cellPadding, NAMESPACE));
    }

    public TableTag<T> withCellSpacing(final CellSpacing cellSpacing) {
        return makeANewOne(attribute("cellspacing", cellSpacing, NAMESPACE));
    }

    public TableTag<T> withFrame(final Frame frame) {
        return makeANewOne(attribute("frame", frame, NAMESPACE));
    }

    public TableTag<T> withRules(final Rules rules) {
        return makeANewOne(attribute("rules", rules, NAMESPACE));
    }

    public TableTag<T> withSummary(final TableSummary tableSummary) {
        return makeANewOne(attribute("summary", tableSummary, NAMESPACE));
    }

    public TableTag<T> withWidth(final XhtmlDimension xhtmlDimension) {
        return makeANewOne(attribute("width", xhtmlDimension, NAMESPACE));
    }
}
